#define _CRT_SECURE_NO_DEPRECATE
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double Cp=2.5;
double Zc=3./8.;
double Ac=9./8.;
double Bc=1./3.;
double Ho=0.;
double So=0.;

double fPr(double Tr,double Vr) /* reduced pressure */
  {
  if(Vr<=Bc)
    return(-999.);
  if(Tr<=0.)
    return(-999.);
  return((Tr/(Vr-Bc)-Ac/Vr/Vr)/Zc);
  }

double fFr(double Tr,double Vr) /* fugacity coefficient */
  {
  double Pr,Z;
  if(Vr<=Bc)
    return(-999.);
  if(Tr<=0.)
    return(-999.);
  Pr=fPr(Tr,Vr);
  if(Pr<=0.)
    return(-999.);
  Z=Zc*Pr*Vr/Tr;
  return(exp(Z-1.-log(Z)+log(Vr/(Vr-Bc))-Ac/Vr/Tr));
  }

double fHr(double Tr,double Vr) /* residual enthalpy */
  {
  double Pr,Z;
  if(Vr<=Bc)
    return(-999.);
  if(Tr<=0.)
    return(-999.);
  Pr=fPr(Tr,Vr);
  if(Pr<=0.)
    return(-999.);
  Z=Zc*Pr*Vr/Tr;
  return(Tr*(1.-Z)+Ac/Vr);
  }

double HofTV(double Tr,double Vr) /* reduced enthalpy */
  {
  return(Ho+Cp*Tr-fHr(Tr,Vr));
  }

double fSr(double Tr,double Vr) /* residual entropy */
  {
  double F,Pr,Hr;
  if(Vr<=Bc)
    return(-999.);
  if(Tr<=0.)
    return(-999.);
  Pr=fPr(Tr,Vr);
  if(Pr<=0.)
    return(-999.);
  F=fFr(Tr,Vr);
  if(F<=0.)
    return(-999.);
  Hr=fHr(Tr,Vr);
  return(Hr/Tr+log(F));
  }

double SofTV(double Tr,double Vr) /* reduced entropy */
  {
  double Pr;
  Pr=fPr(Tr,Vr);
  if(Pr<=0.)
    return(-999.);
  return(So+Cp*log(Tr)-fSr(Tr,Vr)-log(Pr));
  }

typedef struct{double r,i;}COMPLEX;

int Cubic(double*c,COMPLEX*r) /* c0+c1*x+c2*x^2+x^3=0 */
  {
  double a,arm,b,ca,ca3,cb,cb3,d,rm,t,x;
  memset(r,0,3*sizeof(COMPLEX));
  a=c[1]-c[2]*c[2]/3.;
  b=(2.*c[2]*c[2]*c[2]-9.*c[2]*c[1]+27.*c[0])/27.;
  d=b*b/4.+a*a*a/27.;
  if(d>=0.)
    {
    d=sqrt(d);
    ca3=-b/2.+d;
    cb3=-b/2.-d;
    if(ca3<0.)
      ca=-pow(-ca3,1./3.);
    else if(ca3>0.)
      ca=pow(ca3,1./3.);
    else
      ca=0.;
    if(cb3<0.)
      cb=-pow(-cb3,1./3.);
    else if(cb3>0.)
      cb=pow(cb3,1./3.);
    else
      cb=0.;
    r[0].r=ca+cb-c[2]/3.;
    r[1].r=r[2].r=-(ca+cb)/2.-c[2]/3.;
    r[1].i=(ca-cb)*sqrt(3.)/2.;
    r[2].i=-r[1].i;
    return(1);
    }
  rm=2.*sqrt(-a/3.);
  arm=a*rm;
  if(fabs(arm)>0.)
    x=3.*b/arm;
  else
    x=0.;
  t=acos(x)/3.;
  r[0].r=rm*cos(t)-c[2]/3.;
  r[1].r=rm*cos(t+2.*M_PI/3.)-c[2]/3.;
  r[2].r=rm*cos(t+4.*M_PI/3.)-c[2]/3.;
  return(3);
  }

void SatProps(double Tr,double*Pr,double*Vf,double*Vg)
  {
  int i;
  double c[3],P1,P2;
  COMPLEX r[3];
  if(Tr>=1.)
    {
    *Pr=*Vf=*Vg=1.;
    return;
    }
  P1=0.00001;
  P2=1.;
  for(i=0;i<32;i++)
    {
    *Pr=sqrt(P1*P2);
    c[0]=-27.*Pr[0]*Pr[0]/512./Tr/Tr/Tr;
    c[1]=27.*Pr[0]/64./Tr/Tr;
    c[2]=-(Pr[0]/8./Tr+1.);
    if(Cubic(c,r)!=3)
      {
      if(r[0].r>0.5)
        P1=*Pr;
      else
        P2=*Pr;
      continue;
      }
    *Vf=min(r[0].r,min(r[1].r,r[2].r))*Tr/Pr[0]/Zc;
    *Vg=max(r[0].r,max(r[1].r,r[2].r))*Tr/Pr[0]/Zc;
    if(fFr(Tr,*Vf)>fFr(Tr,*Vg))
      P1=*Pr;
    else
      P2=*Pr;
    }
  }

void SaturationTable()
  {
  double Hf,Hg,Ps,Sf,Sg,Ts,Vf,Vg,Zg;
  printf("van der Waals Equation of State\n");
  printf("saturation properties\n");
  printf("Trs   Prs    Vrf     Vrg     Zg    Hrf    Hrg    Srf   Srg\n");
  for(Ts=0.5;Ts<1.001;Ts+=0.01)
    {
    SatProps(Ts,&Ps,&Vf,&Vg);
    Zg=Zc*Ps*Vg/Ts;
    Hf=HofTV(Ts,Vf);
    Hg=HofTV(Ts,Vg);
    Sf=SofTV(Ts,Vf);
    Sg=SofTV(Ts,Vg);
    if(Ho==0.)
      {
      Ho=-Hf;
      So=-Sf;
      Hf+=Ho;
      Hg+=Ho;
      Sf+=So;
      Sg+=So;
      }
    printf("%4.2lf %6.4lf %6.4lf %7.4lf %6.4lf %6.4lf %6.4lf %6.4lf %6.4lf\n",Ts,Ps,Vf,Vg,Zg,Hf,Hg,Sf,Sg);
    }
  }

#define n 20
#define m 25

void MollierChart()
  {
  int i,j;
  double Hr,Ps[m],Sr,Tr,Ts[m],Vf[m],Vg[m],Vr;
  FILE*fp;

  printf("creating Mollier chart\n");
  fp=fopen("waals.csv","wt");

/* subcritical isotherms */

  for(j=0;j<m;j++)
    {
    Tr=0.5+j*0.5/m;
    Ts[j]=Tr;
    SatProps(Tr,Ps+j,Vf+j,Vg+j);
    for(i=0;i<n;i++)
      {
      Vr=Bc*exp((i+1)*log(Vf[j]/Bc)/n);
      Sr=SofTV(Tr,Vr);
      Hr=HofTV(Tr,Vr);
      if(Hr<1.+0.65*Sr)
        fprintf(fp,"%lG,%lG\n",Sr,Hr);
      }
    fprintf(fp,"%lG,%lG\n",SofTV(Tr,Vf[j]),HofTV(Tr,Vf[j]));
    fprintf(fp,"%lG,%lG\n",SofTV(Tr,Vg[j]),HofTV(Tr,Vg[j]));
    for(i=0;i<n;i++)
      {
      Vr=Vg[j]*exp((i+1)*log(100./Vg[j])/n);
      Sr=SofTV(Tr,Vr);
      Hr=HofTV(Tr,Vr);
      fprintf(fp,"%lG,%lG\n",Sr,Hr);
      }
    fprintf(fp,"\n");
    }

/* vapor dome */

  for(j=0;j<m;j++)
    fprintf(fp,"%lG,%lG\n",SofTV(Ts[j],Vf[j]),HofTV(Ts[j],Vf[j]));
  fprintf(fp,"%lG,%lG\n",SofTV(Tr,1.),HofTV(Tr,1.));
  for(j=m-1;j>=0;j--)
    fprintf(fp,"%lG,%lG\n",SofTV(Ts[j],Vg[j]),HofTV(Ts[j],Vg[j]));
  fprintf(fp,"\n");

/* supercritical isotherms */

  for(Tr=1.;Tr<1.61;Tr+=0.02)
    {
    for(i=0;i<n;i++)
      {
      Vr=Bc*exp((i+1)*log(100./Bc)/n);
      Sr=SofTV(Tr,Vr);
      Hr=HofTV(Tr,Vr);
      if(Hr<1.+0.65*Sr)
        fprintf(fp,"%lG,%lG\n",Sr,Hr);
      }
    fprintf(fp,"\n");
    }

  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  SaturationTable();
  MollierChart();
  return(0);
  }
